# $Id: upload.py,v 1.31 2010/11/22 09:13:30 irees Exp $
import sys
import os
import operator
import re
import functools


# emdash imports
import emen2.clients.emdash.models
import emen2.clients.emdash.sessionwizard
import emen2.clients.emdash.controller
import emen2.clients.emdash.ui
import emen2.clients.emdash.config


# PyQt4 imports
from PyQt4 import QtGui, QtCore, Qt

# ian: todo: @abstractmethod decorators in BaseUpload
class BaseUpload(QtGui.QMainWindow):
	# Base class for an Upload UI

	signal_set = QtCore.pyqtSignal(int, unicode, unicode)
	signal_set_setting = QtCore.pyqtSignal(unicode, object)

	def __init__(self, parent=None):
		super(BaseUpload, self).__init__(parent=parent)
		
		# Get a DBProxy handle
		self.db = emen2.clients.emdash.config.getdb()
		
		# Items to display, usually set from controller update signals..
		self.username = None
		self.displayname = None
		self.recids = {}
		self.recnames = {}
		self.comments = {}
		self.path = None		
		self.end = False		
				
		# Setup Controller
		self.controller = self.get_controller()

		# Display model for the queue
		self.queuemodel = emen2.clients.emdash.models.queuemodel.QueueModel(queue=self.controller.queue)

		# Bind controller signals
		self.controller.signal_newgrid.connect(self.queuemodel.newgrid, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_newgrid.connect(self.newgrid, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_update_recs.connect(self.update_recs, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_update_comments.connect(self.update_comments, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_update_setting.connect(self.update_setting, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_begin_session.connect(self.begin_session, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_exception.connect(self.showerror, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_exception.connect(self.login, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_login.connect(self.login_cb, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_file_status.connect(self.queuemodel.file_status, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_file_success.connect(self.queuemodel.file_success, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_file_failure.connect(self.queuemodel.file_failure, type=QtCore.Qt.QueuedConnection)

		self.controller.signal_end_session.connect(self.end_session, type=QtCore.Qt.QueuedConnection)

		# Clock		
		self.controller.clockthread.signal_tick.connect(self.tock, type=QtCore.Qt.QueuedConnection)

		# Update a single item in a record
		self.signal_set.connect(self.controller.set, type=QtCore.Qt.QueuedConnection)
		self.signal_set_setting.connect(self.controller.set_setting, type=QtCore.Qt.QueuedConnection)

		# Setup UI and other slots
		self.ui = self.get_ui()

		# Some basic signals that are common across most Uploaders..
		self.init_common()
		
		# Subclass init
		self.init()


	def init(self):
		# The default Uploader doesn't have any special session management
		self.controller.signal_login.connect(self.controller.begin_no_session, type=QtCore.Qt.QueuedConnection)
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session", self, triggered=self._select_grid_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session by Record ID", self, triggered=self._recid_grid_wizard))
						
			


	def init_common(self):
		# Connect the controller's queue model to self.ui.tree_files
		self.ui.tree_files.setModel(self.queuemodel)
		self.ui.tree_files.setColumnWidth(0, 250)
		self.ui.tree_files.setColumnWidth(1, 60)
		self.ui.tree_files.setColumnWidth(2, 60)
		self.ui.tree_files.setColumnWidth(3, 60)
		self.ui.tree_files.setColumnWidth(4, 60)
		self.ui.tree_files.setColumnWidth(5, 80)
		self.ui.tree_files.setColumnWidth(6, 50)

		# Connect all the newfile events, so the file widget updates
		self.controller.signal_newfile.connect(self.ui.tree_files.newfile, type=QtCore.Qt.QueuedConnection)
		self.controller.signal_newfile.connect(self.queuemodel.newfile, type=QtCore.Qt.QueuedConnection)

		# Connect UI signals back to controller
		self.ui.tree_files.signal_set.connect(self.controller.set, type=QtCore.Qt.QueuedConnection)
		self.ui.tree_files.signal_enqueue.connect(self.controller.uploadthread.enqueue, type=QtCore.Qt.QueuedConnection)
		
		# Drop-down buttons
		self.ui.button_path.addAction(QtGui.QAction("Select directory", self, triggered=self.setpath))

		self.ui.button_session.addAction(QtGui.QAction("Logout", self, triggered=self.controller.end_session))

		self.connect_edits()



	def connect_edits(self):
		pass
		


	def get_controller(self):
		return emen2.clients.emdash.controller.BaseController()



	def get_ui(self):
		u = emen2.clients.emdash.ui.Ui_Upload.Ui_Upload()
		u.setupUi(self)
		return u



	##############################
	# Set / update settings
	##############################

	def set_setting(self, value, param):
		# Change a setting, usually sends to controller
		# print "upload: set_setting: %s %s"%(value, param)
		if param:
			self.signal_set_setting.emit(param, value)
			
				

	def update_ui(self, comments=False):
		# Update the UI from a Record change in the controller.
		self.ui.label_grid.setText(self.getlabel(["grid_imaging"]))


		
	def update_setting(self, param, value):	
		# Update a UI setting, usually from controller
		pass
			
			

	@QtCore.pyqtSlot(object, object, object)
	def update_recs(self, recids, recs, recnames):
		# Controller saved some records, callback to the UI to reflect this
		self.recids = recids
		self.recs = recs
		self.recnames = recnames
		self.update_ui()
				
				
	
	@QtCore.pyqtSlot(object)
	def update_comments(self, l):
		# Callback for updated comments
		self.comments = l
		self.update_ui(comments=True)
		


	@QtCore.pyqtSlot(str)
	def update_path(self, path):
		self.ui.label_path.setText(unicode(path))



	@QtCore.pyqtSlot(int)
	def newgrid(self, newgrid):
		if newgrid > -1:
			self.ui.button_path.setEnabled(True)
			self.ui.frame_settings.setEnabled(True)
			try: self.ui.button_newmicrograph.setEnabled(True)
			except: pass
		else:
			self.ui.button_path.setEnabled(False)
			self.ui.frame_settings.setEnabled(False)
			try: self.ui.button_newmicrograph.setEnabled(False)
			except: pass


	##############################
	# Session
	##############################


	@QtCore.pyqtSlot()
	@QtCore.pyqtSlot(unicode)
	@QtCore.pyqtSlot(unicode, unicode, unicode)
	def login(self, username=None, password=None):
		# Get login info for new session
		authmsg = "Login to %s"%emen2.clients.emdash.config.options.host
		l = Login(parent=self, authmsg=authmsg, username=username, password=password)
		l.signal_getlogin.connect(self.controller.login)
		if username and password:
			l.signal_getlogin.emit(username, password)
		else:
			l.show()



	@QtCore.pyqtSlot(unicode, unicode, unicode)
	def login_cb(self, ctxid, username, displayname):
		# Callback for successful login
		self.db._setContext(ctxid)
		self.username = username
		self.displayname = displayname
		self.show()



	@QtCore.pyqtSlot(int)
	def begin_session(self, session):
		# Callback for session begin
		if session > -1:
			self.recids[emen2.clients.emdash.config.options.session_rectype] = session

		if self.username:
			self.ui.label_user.setText("%s (%s)"%(self.displayname, self.username))

		self.update_ui()
		self.show()


		
	@QtCore.pyqtSlot()	
	def end_session(self, restart=True):
		# End session
		if self.end:
			return
			
		print "Ending session..."			
		s = self.recids.get(emen2.clients.emdash.config.options.session_rectype)
		self.end = True
		self.close()
		t = DWarning(parent=self, displayname=self.displayname, session=s)
		self.controller.finished.connect(t.accept)
		t.show()
	
	
			
	################
	# Clock
	################
	
	@QtCore.pyqtSlot(str)	
	def tock(self, t):
		# Update session clock
		target = self.recids.get(emen2.clients.emdash.config.options.session_rectype)		
		if t == None:
			t = ""
		elif target > -1:
			t = """<a href="%s/record/%s/?ctxid=%s">%s</a>"""%(emen2.clients.emdash.config.options.host, target, self.db._ctxid, t)	
		else:
			t = """<a href="%s?ctxid=%s">%s</a>"""%(emen2.clients.emdash.config.options.host, self.db._ctxid, t)	
			
		self.ui.label_session.setText(t)



	#####################
	# Event handlers
	#####################
	
	def closeEvent(self, event):
		# event.ignore()
		self.controller.end_session()
	
		
			
	@QtCore.pyqtSlot(str)
	def showerror(self, e):
		error = QtGui.QErrorMessage(self)
		error.showMessage(e)
		error.exec_()		



	#####################
	# Comments
	#####################

	@QtCore.pyqtSlot(unicode)
	def addcomment(self, rectype, recid=None):
		# Add a comment to a known record
		comment = self.get_comment_text()
		if rectype:
			recid = int(self.recids.get(rectype))
		self.signal_set.emit(recid, "comments", comment)



	def get_comment_text(self):
		# Get comment text and do any UI updates to reflect this, e.g. clear text field..
		return ""



	#####################
	# Signals to FS Poll Thread
	#####################
	
	@QtCore.pyqtSlot()
	def setpath(self):
		# Select directory to watch for queue
		d = QtGui.QFileDialog(self)
		d.setFileMode(QtGui.QFileDialog.Directory)
		d.setDirectory("/")
		d.fileSelected.connect(self.controller.fspollthread.setpath, type=QtCore.Qt.QueuedConnection)
		d.fileSelected.connect(self.update_path)		
		d.exec_()
		
		
		
	@QtCore.pyqtSlot()
	def addfiles(self):
		# Directly add some files to the queue
		pass
		# d = QtGui.QFileDialog(self)
		# d.fileSelected.connect(self.controller.fspollthread.setpath, type=QtCore.Qt.QueuedConnection)
		# d.exec_()
			
			
			
	#####################
	# Misc.
	#####################
			
	def getlabel(self, rectypes):
		"""return the recid/recname link for the first rectype found"""
		if not hasattr(rectypes, "__iter__"):
			rectypes = [rectypes]
			
		for r in rectypes:
			if self.recids.get(r) != None:
				return """<a href="%s/record/%s?ctxid=%s">%s</a>"""%(emen2.clients.emdash.config.options.host, self.recids.get(r), self.db._ctxid, self.recnames.get(str(self.recids.get(r)), r))
		return ""



	def connect_slider(self, i, factor=1000.0):
		slider = getattr(self.ui, 'slider_%s'%i)
		edit = getattr(self.ui, 'edit_%s'%i)
		slider.sliderMoved.connect(functools.partial(self._slider_to_edit, target=edit, factor=factor), type=QtCore.Qt.QueuedConnection)
		edit.textChanged.connect(functools.partial(self._edit_to_slider, target=slider, factor=factor), type=QtCore.Qt.QueuedConnection)
		edit.textChanged.connect(functools.partial(self.set_setting, param=i), type=QtCore.Qt.QueuedConnection)		

	def _slider_to_edit(self, i, target=None, factor=1.0):
		i = i / factor
		target.setText(str(i))

	def _edit_to_slider(self, i, target=None, factor=1.0):
		i = float(i) * factor
		target.setValue(int(float(i or 0)))	



	#####################
	# Default wizards..
	#####################

	def _select_grid_wizard(self):
		# Edit the microscopy session we just created to fill in details
		wizard = emen2.clients.emdash.sessionwizard.SelectGridWizard(parent=self, selected=self.recids, db=emen2.clients.emdash.config.getdb(ctxid=self.db._ctxid))
		wizard.signal_done.connect(self.controller.newgrid, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()



	def _recid_grid_wizard(self):
		# d, ok = QtGui.QInputDialog.getInt(self, "Existing Grid Imaging", "Record ID")
		d = QtGui.QInputDialog(self)
		d.setLabelText("Record ID")
		d.setIntMaximum(2147483647)
		d.setIntMinimum(0)
		d.setInputMode(QtGui.QInputDialog.IntInput)
		d.intValueSelected.connect(self.controller.selectgrid)
		d.exec_()








	

class MicroscopeUpload(BaseUpload):
	
	def init(self):
		# Comments
		self.ui.button_addcomment.addAction(QtGui.QAction("... this grid", self, triggered=functools.partial(self.addcomment, rectype='grid_imaging')))
		self.ui.button_addcomment.addAction(QtGui.QAction("... this microscopy session", self, triggered=functools.partial(self.addcomment, rectype='microscopy')))
		self.ui.button_addcomment.addAction(QtGui.QAction("... this microscope", self, triggered=functools.partial(self.addcomment, rectype='microscope')))			

		# This Uploader needs to collect additional data from the user before a session can be started
		self.controller.signal_login.connect(self._newmicroscopy_wizard, type=QtCore.Qt.QueuedConnection)

		# More Drop-down buttons
		self.ui.button_grid.addAction(QtGui.QAction("New Grid Imaging Session", self, triggered=self._newgrid_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("New Freezing Session", self, triggered=self._newfreezing_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("Edit Grid Imaging Session Details", self, triggered=self._edit_grid_imaging))
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session", self, triggered=self._select_grid_wizard))
		self.ui.button_grid.addAction(QtGui.QAction("Select Existing Grid Imaging Session by Record ID", self, triggered=self._recid_grid_wizard))

		self.ui.button_session.addAction(QtGui.QAction("Edit Session", self, triggered=self._edit_microscopy))
			
			

	def get_ui(self):
		u = emen2.clients.emdash.ui.Ui_MicroscopeUpload.Ui_Upload()
		u.setupUi(self)
		self.ui = u
		self._toggle_sessiontype()
		return u	



	def get_controller(self):
		return emen2.clients.emdash.controller.MicroscopeController()		
			

				
	def update_setting(self, param, value):
		try:
			if param == "id_micrograph":
				self.ui.edit_id_micrograph.setText(value)
	
			elif param == "ctf_defocus_set":
				pass
	
			elif param == "tem_magnification_set":
				pass

		except:
			pass
			


	def update_ui(self, comments=False):
		# Update the UI to reflect these selected records
		self.ui.label_grid.setText(self.getlabel(["grid_imaging"]))		
		self.ui.label_project.setText(self.getlabel(["project","subproject"]))
		self.ui.label_freeze.setText(self.getlabel(["grid_preparation"]))
		
		film_type = self.recs.get(self.recids.get(emen2.clients.emdash.config.options.session_rectype), dict()).get('film_type', '')
		print "Checking film_type:", film_type
		if film_type.lower() in emen2.clients.emdash.config.film_types:
			self._toggle_sessiontype(session='film')

		self._update_controls()

		if comments:
			self._update_ui_comments()



	def _update_ui_comments(self):
		html = []	
		sortlist = []
		for k,v in self.comments.items():
			for c in v:
				sortlist.append([k]+c)
								
		sortlist = sorted(sortlist, key=operator.itemgetter(2), reverse=True)		
				
		for c in sortlist:
			t = """<a href="%s/record/%s/?ctxid=%s">%s</a>: %s @ %s: <br /> %s <br /><br />"""%(emen2.clients.emdash.config.options.host, c[0], self.db._ctxid, self.recnames.get(c[0], c[0]), c[1], c[2], c[3])
			html.append(t)
				
		self.ui.browser_comments.setHtml("".join(html))
				
				


	def get_comment_text(self):
		text = self.ui.edit_addcomment.toPlainText()		
		self.ui.edit_addcomment.clear()
		return text



	#####################
	# Wizards!!
	#####################


	@QtCore.pyqtSlot()
	def _newmicroscopy_wizard(self):
		wizard = emen2.clients.emdash.sessionwizard.NewMicroscopyWizard(parent=self, selected=self.recids, db=emen2.clients.emdash.config.getdb(ctxid=self.db._ctxid))
		wizard.signal_done.connect(self.controller.begin_session, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()


	@QtCore.pyqtSlot()
	def _newfreezing_wizard(self):
		wizard = emen2.clients.emdash.sessionwizard.NewFreezingWizard(parent=self, selected=self.recids, db=emen2.clients.emdash.config.getdb(ctxid=self.db._ctxid))
		wizard.signal_done.connect(self.controller.newfreezing, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()



	@QtCore.pyqtSlot()
	def _newgrid_wizard(self):
		wizard = emen2.clients.emdash.sessionwizard.NewGridWizard(parent=self, selected=self.recids, db=emen2.clients.emdash.config.getdb(ctxid=self.db._ctxid))
		wizard.signal_done.connect(self.controller.newgrid, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()



	def _existingmicroscopy_wizard(self):
		# d, ok = QtGui.QInputDialog.getInt(self, "Existing Grid Imaging", "Record ID")
		d = QtGui.QInputDialog(self)
		d.setLabelText("Record ID")
		d.setIntMaximum(2147483647)
		d.setIntMinimum(0)
		d.setInputMode(QtGui.QInputDialog.IntInput)
		d.intValueSelected.connect(self.controller.selectmicroscopy)
		d.exec_()



	def _edit_microscopy(self):
		# Edit the microscopy session we just created to fill in details
		wizard = emen2.clients.emdash.sessionwizard.EditMicroscopyWizard(parent=self, selected=self.recids, db=emen2.clients.emdash.config.getdb(ctxid=self.db._ctxid))
		wizard.signal_done.connect(self.controller.edit_records, type=QtCore.Qt.QueuedConnection)
		wizard.exec_()		



	def _edit_grid_imaging(self):
		if self.recids.get('grid_imaging') != None:
			wizard = emen2.clients.emdash.sessionwizard.EditGridImagingWizard(parent=self, selected=self.recids, db=emen2.clients.emdash.config.getdb(ctxid=self.db._ctxid))
			wizard.signal_done.connect(self.controller.edit_records, type=QtCore.Qt.QueuedConnection)
			wizard.exec_()



	##############################
	# Control handlers...
	#############################

	def connect_edits(self):		
		# New Micrograph
		self.ui.button_newmicrograph.clicked.connect(self.controller.newmicrograph, type=QtCore.Qt.QueuedConnection)			

		# Editables
		self.connect_slider('tem_magnification_set', factor=1.)
		self.connect_slider('ctf_defocus_set')
		self.connect_slider('tem_dose_rate')
		self.connect_slider('time_exposure_tem')
		
		self.ui.edit_id_micrograph.textChanged.connect(functools.partial(self.set_setting, param='id_micrograph'), type=QtCore.Qt.QueuedConnection)
		self.ui.button_assess_ice_comments.clicked.connect(self._assess_ice_comments)
		self.ui.button_assess_ice_thick.clicked.connect(self._assess_ice_thick)



	def _toggle_sessiontype(self, session=None):
		if session != 'film':
			self.ui.edit_id_micrograph.hide()
			self.ui.label_newmicrograph.hide()
			self.ui.button_newmicrograph.hide()		
		else:
			self.ui.edit_id_micrograph.show()
			self.ui.label_newmicrograph.show()
			self.ui.button_newmicrograph.show()			



	def _update_controls(self):
		gir = self.recs.get(self.recids.get("grid_imaging"), dict())
		if gir.get('tem_magnification_set') != None:
			self.ui.edit_tem_magnification_set.setText(str(gir.get('tem_magnification_set')))
		if gir.get('ctf_defocus_set') != None:
			self.ui.edit_ctf_defocus_set.setText(str(gir.get('ctf_defocus_set')))
		if gir.get('time_exposure_tem') != None:
			self.ui.edit_time_exposure_tem.setText(str(gir.get('time_exposure_tem')))
		if gir.get('tem_dose_rate') != None:
			self.ui.edit_tem_dose_rate.setText(str(gir.get('tem_dose_rate')))


		
	################
	# Comments about ice quality
	################
	
	def _assess_ice_thick(self):
		value = unicode(self.ui.assess_ice_thick.currentText())
		self.signal_set.emit(self.recids.get("grid_imaging"), "assess_ice_thick", value)

		
		
	def _assess_ice_comments(self):
		value = unicode(self.ui.assess_ice_comments.currentText())
		self.signal_set.emit(self.recids.get("grid_imaging"), "assess_ice_comments", value)
		# trivial little hookups between controls..






# This handler is basically the same as the Microscopy handler, but simpler in a few places.
class ScanUpload(BaseUpload):
	def get_ui(self):
		u = emen2.clients.emdash.ui.Ui_ScanUpload.Ui_Upload()
		u.setupUi(self)
		return u		
		

	##############################
	# Control handlers...
	#############################

	def connect_edits(self):		
		# Editables
		self.connect_slider('scan_step')
		self.connect_slider('nikon_gain')
		self.connect_slider('scan_average', factor=10.0)
		self.connect_slider('angstroms_per_pixel')		

		self.ui.edit_scanner_film.editTextChanged.connect(functools.partial(self.set_setting, param='scanner_film'), type=QtCore.Qt.QueuedConnection)
		self.ui.edit_scanner_cartridge.editTextChanged.connect(functools.partial(self.set_setting, param='scanner_cartridge'), type=QtCore.Qt.QueuedConnection)
		







##############################
# Some dialogs: Warning, Login, ...
##############################


class DWarning(QtGui.QDialog):
	
	def __init__(self, displayname=None, session=None, parent=None):
		QtGui.QDialog.__init__(self, parent=parent)
		
		self.ui = emen2.clients.emdash.ui.Ui_Warning.Ui_Warning()
		self.ui.setupUi(self)
		self.ui.label_user.setText(unicode(displayname))
		self.ui.label_session.setText("""<a href="%s/record/%s/">%s</a>"""%(emen2.clients.emdash.config.options.host, session, session))

	def closeEvent(self, event):
		event.ignore()




class Login(QtGui.QDialog):
	
	signal_getlogin = QtCore.pyqtSignal(unicode, unicode)	
	
	def __init__(self, username=None, password=None, authmsg=None, parent=None):
		QtGui.QDialog.__init__(self, parent=parent)		
		self.ui = emen2.clients.emdash.ui.Ui_Login.Ui_Login()
		self.ui.setupUi(self)

		if authmsg:
			self.ui.label_login.setText(authmsg)
		if username:
			self.ui.edit_username.setText(username)
		if password:
			self.ui.edit_password.setText(password)
		
			
	# turn on before deployment..
	def reject(self):
		return


	def accept(self):
		if not self.ui.edit_username.text():
			return
		self.signal_getlogin.emit(unicode(self.ui.edit_username.text()), unicode(self.ui.edit_password.text()))
		self.done(0)








__version__ = "$Revision: 1.31 $".split(":")[1][:-1].strip()
